/*

DUMMY.H - header file that declares classes, objects, and functions
          not otherwise included in this group of files

This work is dedicated to the public domain by the author,
Edward J. Lansinger (lansie@rpi.edu).  It is therefore not
protected by copyright and can be used freely by anyone
for any purpose.

Some source files have been left out of this archive, as they
do not pertain directly to LCD display classes.  Including them
is an unnecessary burden, as to do so would bring in large chunks
of a data acquisition and control system without further illuminating
how to set up classes to control an LCD.

Instead, those declarations that are necessary for a clean compile
are included here.  A clean link is not possible without the actual
definitions, of course; this is left as an exercise for the reader.


*/

#ifndef _DUMMY_H
#define _DUMMY_H

#include "std.h"

class ProcessClass
// ProcessClass objects are called by a round-robin scheduler
// (not defined in this archive).  Each object has a Go() function
// that is called non-preemptively by the scheduler.  The Go()
// function is expected to do a short burst of work and then
// return.
{
  BYTE *HeapPtr;
  UINT HeapSize;
  ProcessClass *pNextProcessInList;
protected:
  ProcessClass *GetLastProcessAddedToList();
  virtual void Go();
public:
  ProcessClass();
};

/*

Datalogging

Any variable in the code that the programmer wishes to log is derived
from the class PRMClass (for "PaRaMeter").  PRMClass takes care
of recording useful information for all recordable parameters, such
as each one's displayable name and unit of measure.  The DataLogger (below)
class/object operates on the list of PRMClass objects that is created
as each such object is instantiated.

*/

enum SampleSizes {IgnoredSample, BitSample, ByteSample, WordSample, DWordSample, intSample};

UINT BytesPerSample(SampleSizes s);

class PRMClass
{
  PRMClass *PrevPRM;
  PRMClass *NextPRM;
public:
  PRMClass();
  PRMClass *Last(); // return last PRM added
  PRMClass *Next(); // return next in (circular) list after p
  PRMClass *Prev(); // return prev in (circular) list before p
  virtual SampleSizes GetSampleSize(); // returns IgnoredSample if to be ignored
  virtual char *GetName(); // i.e. "YAW RATE"
  virtual char *GetUnits(); // i.e. "DEG/SEC"
  virtual DWORD GetSample(); // returns a sample
  virtual UINT GetScaledValue(DWORD S,
                              UINT Max
                             ); // returns S (a value
          // generated by GetSample()) scaled
          // to fit in the range [0, Max]; useful for graphing
  virtual void GetDisplayValue(DWORD S,
                               char *Buf
                              ); // formats a value for display
          // into the text buffer Buf from the value S which
          // should be generated by GetSample()
          // The output string has a constant length, so redrawing
          // a new value at the same place on the screen will blot
          // out the old value.
  virtual char *GetDisplayUnits(); // returns a pointer to text
          // describing the units appropriate to the value
          // returned by GetDisplayValue to be used for display purposes
};

extern PRMClass *LastPRMAddedToList;
extern UINT SampleablePRMCount;

/*

Here's how data is logged:

Every sample interval, all objects derived from PRMClass
are queried for a value.  This value is returned in a DWORD, but may
not fill the entire DWORD.  The PRMClass object identifies just how
many bytes are used.  Anyways, the PRMClass objects are queried in
the order of the master PRMClass list for samples, and these samples
are concatenated (using 1, 2, 3, or 4 bytes per sample per the actual
size) in the same order and placed in a temporary array.

Each time a new flight recording finishes, an object of type DataSet
is created and added to a list of DataSet objects.  An array is
created in far memory to which the contents of the temporary array
are copied.  This is the array that the PC software can access.

The DataLogger is controlled by the IOProcess class so that it is
active regardless of what is being displayed on the screen.

*/

struct SampleInfo
{
  UINT      Size;
  BYTE  *Name; // this is stored here so the PC can get it without having
  BYTE  *Units; // to talk to the PRMClass that created it
  PRMClass *P; // pointer to the PRMClass that created the sample
};

struct DataSet
{
  UINT NumRecs; // number of records in this set
  UINT RecSize; // number of bytes per record
  UINT SamplesPerRec;
  SampleInfo *Info;
  BYTE *Recs; // array of agglomerated samples (stored in far memory,
              // possibly compressed)
  PRMClass *PRMs; // array of pointers to the parameters that created the
                  // samples
  DataSet *next; // next data set in list
};

enum TriggerTypes {StartTrigger, EndTrigger};

class DataLoggerClass
{
  BOOL Recording;
  UINT RecCount; // number of records recorded so far
  UINT MaxRecCount;
  BOOL StopRecording; // set to !FALSE when recording is to stop
  UINT PRMCount; // number of Parameters to record
  UINT RecSize;
  void *TmpStorage;
  BYTE *TmpHead;
  BYTE *TmpTail;
  BYTE *LastRecInTmp;
  TriggerTypes Trig;
  UINT DataSetCount;
public:
  DataLoggerClass();
  BOOL SetRecordTime(UINT s); // sets recording time in seconds, returns
                              // TRUE if recording is possible
  BOOL Enable(TriggerTypes t); // returns FALSE if cannot record
  void Trigger();
  void Stop(); // stops recording and saves data set
  UINT GetNumDataSets(); // returns number of data sets stashed
  BOOL IsRecording(); // returns FALSE if not recording
private:
  void GetSamples(); // called every sample period
  void StashSamples(); // stash where they are safe and PC can find them
};

extern DataSet *LastDataSet; // points to a linked list of stashed
                             // data sets

extern DataLoggerClass DataLogger;

/*

Here's a simple timepiece, update by a system interrupt (not shown).

*/

class SlowClockClass
{
public:
  SlowClockClass();
  ULONG Time(); // returns time since last RESET in thirty-seconds of a second
};

extern SlowClockClass SlowClock;

/*

Input

The original data acquisition and control system was designed to be used
by the driver of a race car.  Three major input devices were defined:
two buttons on the steering wheel, and a dial-type potentiometer
on the dashboard.

*/

/* Steering wheel buttons */
enum SWButtons {RightSWB, LeftSWB};

class SWButton : ProcessClass
{
/* Designed for press-on/press-off switches */
protected:
//  InputBit *Bit;
  UINT History; // used for debouncing
  BOOL State;
  UINT PressCount;
  virtual void Go();
public:
  SWButton(SWButtons B);
  UINT GetPressCount();
    /* returns number of times button was pressed since last call
    to this routine, up to 255 */
};

class DialClass : PRMClass
{
//  ADCChannel A;
public:
  DialClass();
  UINT GetVal1024(); // returns dial position 0..1023
protected:
  virtual SampleSizes GetSampleSize();
  virtual char *GetName();
  virtual char *GetUnits();
  virtual DWORD GetSample();
  virtual UINT GetScaledValue(DWORD S, UINT Max);
  virtual void GetDisplayValue(DWORD S, char *Buf);
  virtual char *GetDisplayUnits();
};

extern SWButton RightSWButton(RightSWB);
extern SWButton LeftSWButton(LeftSWB);
extern DialClass Dial;


#endif
